/**
 * Copyright 2014 (c) Baidu, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
#ifndef BAIDU_BOS_CPPSDK_TAG_CLIENT_H
#define BAIDU_BOS_CPPSDK_TAG_CLIENT_H

#include <stdint.h>

#include <string>
#include <map>
#include <functional>

#include "bcesdk/bos/client_options.h"
#include "bcesdk/common/common.h"
#include "bcesdk/auth/auth.h"
#include "bcesdk/http/http_client.h"
#include "bcesdk/tag/list_tag_request.h"
#include "bcesdk/tag/list_tag_response.h"


BEGIN_CPPSDK_NAMESPACE

class TagClientImpl {
public:
    TagClientImpl(const Credential &credential, const ClientOptions &options): _options(options) {

        DefaultSigner *signer = new DefaultSigner(options);
        signer->set_credential(credential);
        _signer = signer;
        if (_options.user_agent.empty()) {
            _options.user_agent = sdk_package_string();
        }
    }
    virtual ~TagClientImpl() {
        
       if (_signer != NULL) {
           delete _signer;
       }
    }

    const ClientOptions &options() const { 
        return _options; 
    }

    int build_http_request(BceRequest &request, HttpRequest *http_request) {
        http_request->set_endpoint(_options.endpoint, _options, request.bucket_name());
        http_request->set_timeout(_options.timeout);
        http_request->set_connect_timeout_ms(_options.connect_timeout_ms);
        http_request->append_header("Host", http_request->host());
        http_request->append_header("User-Agent", _options.user_agent);
        int ret = request.build_http_request(http_request);
        if (ret != 0) {
            return ret;
        }
        http_request->append_header("x-bce-date", TimeUtil::now_utctime());
        _signer->sign(http_request);
        return 0;
    }

    int send_request(BceRequest &request, BceResponse *response) {

        HttpRequest http_request;
        int ret = build_http_request(request, &http_request);
        if (RET_OK != ret) {
            LOG(ERROR) << "build http request failed, ret:" << ret;
            return ret;
        }
        HttpResponse *http_response = response->mutable_http_response();
        int try_num = 1;
        if (_options.retry > 0) {
            try_num += _options.retry;
        }
        HttpClient client;
        while (try_num-- > 0) {
            ret = client.execute(http_request, http_response);
            if (ret == 0) {
                response->handle_response(*http_response);
            } else {
                response->set_error(STATUS_RPC_FAIL, "HttpExecuteFailure");
                LOG(WARN) << "http execute error: (" << ret << ')' << stringfy_ret_code(ret);
            }
            if (response->is_ok()) {
                return 0;
            }
            ret = response->status_code();
            if (ret >= 500 && ret != 501) {
                if (try_num > 0) {
                    InputStream *istream = http_request.get_input_stream();
                    if (istream != NULL) {
                        istream->seek(0);
                    }
                    http_response->reset();
                    continue;
                }
            }
            break;
        }
        return ret;
    }

private:
    ClientOptions _options;
    Signer *_signer;
};


class TagClient {
public:
    TagClient(const std::string &access_key_id, const std::string &access_key_secret) {
        ClientOptions option;
        option.endpoint = "bj.bcebos.com";
        _impl = new TagClientImpl(Credential(access_key_id, access_key_secret), option);
    }
    TagClient(const std::string &access_key_id, const std::string &access_key_secret,
            const ClientOptions &options) {
        _impl = new TagClientImpl(Credential(access_key_id, access_key_secret), options);
    }
    TagClient(const Credential &credential, const ClientOptions &options) {
        _impl = new TagClientImpl(credential, options);
    }

    ~TagClient() {
        if (_impl) {
            delete _impl;
            _impl = NULL;
        }
    }

    int list_tag(ListTagRequest &request, ListTagResponse *result) const {
        return _impl->send_request(request, result);
    }
private:
    TagClientImpl *_impl;
};

END_CPPSDK_NAMESPACE
#endif

