/**
 * bosfs - A fuse-based file system implemented on Baidu Object Storage(BOS)
 *
 * Copyright (c) 2016 Baidu.com, Inc. All rights reserved.
 *
 * @file    bosfs_util.h
 * @author  Song Shuangyang(songshuangyang@baidu.com)
 * @date    2016.9
 **/
#ifndef BAIDU_BOS_BOSFS_BOSFS_UTIL_H
#define BAIDU_BOS_BOSFS_BOSFS_UTIL_H

#include <stdint.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>

#include <string>
#include <map>

#include "common.h"
#include "util.h"
#include "data_cache.h"
#include "bcesdk/bos/client.h"
#include "bcesdk/bos/response/list_objects_response.h"
#include "bcesdk/bos/response/head_bucket_response.h"

BEGIN_FS_NAMESPACE

using baidu::bos::cppsdk::ObjectMetaData;
using baidu::bos::cppsdk::BucketType;

class DataCacheEntity;

// Directory type
enum dir_type_t {
    DIR_UNKNOWN  = -1,
    DIR_NEW      = 0,
    DIR_OLD      = 1,
    DIR_FOLDER   = 2,
    DIR_NOOBJ    = 3
};

enum StatConstraint {
    ST_BLKSIZE = 4096,
    ST_BLOCKSIZE = 512,
    ST_MINBLOCKS = 8
};

void die(const char *format, ...);

class BosfsUtil{
private:
    class CaseInsensitiveComp {
    public:
        bool operator()(const std::string &a, const std::string &b) const {
            return strcasecmp(a.c_str(), b.c_str()) < 0;
        }
    };

public:
    typedef std::map<std::string, std::string, CaseInsensitiveComp> mimes_t;

public:
    static SharedPtr<baidu::bos::cppsdk::Client> bos_client();
    static int create_bos_client();

    static int reload();
    static void *reload_thread(void *param);
    static void start_reload();
    static void stop_reload();
    static void reload_sighandler(int sig);

    static int load_credentials();

    static int init_bos();

    static void set_log_level(const std::string &level);
    static void set_log_file(const std::string &file);
    static void set_log_rotate_size(int rotate_size);

    static inline bool is_replace_dir(int type) {
        return type == DIR_OLD || type == DIR_FOLDER || type == DIR_NOOBJ;
    }
    static inline bool is_empty_dir(int type) {
        return type == DIR_OLD || type == DIR_FOLDER;
    }
    static inline bool is_dir_path(const std::string &path) {
        return path[path.length() - 1] == '/';
    }
    static std::string object_to_path(const std::string &object);
    static std::string object_to_basename(const std::string &object, const std::string &prefix);

    static std::string get_username(uid_t uid);
    static std::string get_real_path(const char*);
    static int is_uid_in_group(uid_t uid, gid_t gid);
    static std::string bosfs_basename(const char *path);
    static std::string bosfs_basename(std::string path);

    static int mkdirp(const std::string &path, mode_t mode);
    static bool check_exist_dir_permission(const char *dirpath);
    static bool delete_files_in_dir(const char *dir, bool is_remove_own);

    static bool init_mimetype(const char *mime_file=NULL);
    static std::string get_mimetype(const std::string &path_name);

    static bool set_mountpoint_attr(struct stat &ps);
    static int check_object_access(const char *path, int mask, struct stat *pstbuf);
    static int get_object_attribute(const std::string &path, struct stat *pstbuf,
            ObjectMetaData *pmeta = NULL);
    static int check_path_accessible(const char *path);
    static int check_parent_object_access(const char *path, int mask);
    static int check_object_owner(const char *path, struct stat *pstbuf);
    static DataCacheEntity *get_local_entity(const char *path, bool is_load=false);

    static int exist_bucket();
    static int exist_bucket_mock();
    static int head_bucket();
    static int create_bucket();
    static int check_bucket_access();
    static int head_object(const std::string &object, ObjectMetaData *meta, bool *is_dir_obj,
            bool *is_prefix);
    static int multiple_head_object(std::vector<std::string> &objects,
            std::vector<struct stat *> &stats);

    static int list_subitems(const std::string &prefix, int max_keys,
            std::vector<std::string> *items) {
        std::string marker;
        return list_objects(prefix, max_keys, marker, "/", items);
    }
    // if common_prefix vector not given, then common prefixes will put into items vector
    static int list_objects(const std::string &prefix, int max_keys, std::string &marker,
            const char *delimiter, std::vector<std::string> *items,
            std::vector<std::string> *common_prefix = NULL,
            std::vector<struct stat> *stats = NULL);

    static void create_meta(const std::string &object_name, mode_t mode, uid_t uid, gid_t gid,
            ObjectMetaData *meta);
    static int create_object(const char *path, mode_t mode, uid_t uid, gid_t gid);
    static int create_object(const char *path, mode_t mode, uid_t uid, gid_t gid,
            const std::string &data);
    static int delete_object(const std::string &object, std::string *version=NULL);

    static int rename_file(const std::string &path, const std::string &dst, int64_t size_hint = -1);
    static int rename_directory(const std::string &src, const std::string &dst);

    static int change_object_meta(const std::string &object, ObjectMetaData &meta);

    static void init_default_stat(struct stat *pst);
    static void init_stat_from_listobject(struct stat *pst, const baidu::bos::cppsdk::ObjectSummary& object_summary);

private:
    static SharedPtr<baidu::bos::cppsdk::Client> _s_bos_client;
    static const char * _s_default_mime_file;
    static mimes_t _s_mime_types;
};

END_FS_NAMESPACE

#endif

