/**
 * bosfs - A fuse-based file system implemented on Baidu Object Storage(BOS)
 *
 * Copyright (c) 2016 Baidu.com, Inc. All rights reserved.
 *
 * @file    data_cache.h
 * @author  Song Shuangyang(songshuangyang@baidu.com)
 * @date    2016.9
 **/
#ifndef BAIDU_BOS_BOSFS_DATA_CACHE_H
#define BAIDU_BOS_BOSFS_DATA_CACHE_H

#include <stdint.h>       // for standard int type
#include <sys/types.h>    // for system types(off_t...)
#include <sys/statvfs.h>  // for statvfs/fstatvfs
#include <sys/stat.h>     // for struct stat
#include <stdio.h>        // for standard io of FILE

#include <string>
#include <vector>
#include <map>
#include <list>

#include <pthread.h>

#include "common.h"
#include "mem_buf.h"
#include "object_page.h"
#include "util.h"
#include "bcesdk/bos/client.h"

#if defined(P_tmpdir)
#define TEMP_FILE_DIR P_tmpdir
#else
#define TEMP_FILE_DIR "/tmp"
#endif

BEGIN_FS_NAMESPACE

using baidu::bos::cppsdk::ObjectMetaData;

class StatCacheFile {
public:
    static bool delete_file(const char *path);
    static bool check_top_dir();
    static bool delete_dir();

    explicit StatCacheFile(const char *path = NULL);
    ~StatCacheFile();

    bool open_file();
    bool release();
    const std::string& get_path() const { return _path; }
    bool set_path(const char *path, bool is_open = true);
    int get_fd() const {
        return _fd;
    }

private:
    static bool make_path(const char *path, std::string &file_path, bool is_create_dir = true);

private:
    std::string _path;
    int         _fd;
};

class DataCacheEntity {
public:
    typedef baidu::bos::cppsdk::MultipleUploadHelper MultiUploadHelper;

    DataCacheEntity(const char *tpath = NULL, const char *cpath = NULL);
    ~DataCacheEntity();

    int close_file();
    int open_file(ObjectMetaData *pmeta = NULL, ssize_t size = -1, time_t time = -1);
    bool open_and_load_all(ObjectMetaData *pmeta = NULL, size_t *size = NULL, bool force_load = false);
    int dup_file();

    bool is_open() const {
        return -1 != _fd;
    }
    bool is_tmpfile() const {
        return _is_tmpfile;
    }
    const char *get_path() const {
        return _path.c_str();
    }
    void set_path(const std::string &newpath) {
        _path = newpath;
    }
    void set_modified(bool is_modified) {
        _is_modified = is_modified;
    }
    int get_fd() const {
        return _fd;
    }
    bool is_use_mem_cache() {
        return _use_mem_cache;
    }
    bool set_is_mem_cache(bool is_use) {
        _use_mem_cache = is_use;
    }

    ERwStrategy rw_strategy() const {
        return _rw_strategy;
    }

    void set_rw_strategy(ERwStrategy strategy) {
        _rw_strategy = strategy;
    }

    bool is_safe_disk_space(size_t size);
    bool get_stats(struct stat &st);
    int set_mtime(time_t time);
    bool update_mtime();
    bool get_size(size_t &size);
    bool set_mode(mode_t mode);
    bool set_uid(uid_t uid);
    bool set_gid(gid_t gid);
    bool set_content_type(const char *path);
    void set_xattr(const std::string &xattr);

    int load(off_t start = 0, size_t size = 0);
    int row_flush(const char *tpath, bool force_sync = false);
    int flush(bool force_sync = false);
    ssize_t read(char *bytes, off_t start, size_t size, bool force_sync = false);
    ssize_t write(const char *bytes, off_t start, size_t size);

    int truncate(off_t size);

private:
    static int fill_file(int fd, unsigned char byte, size_t size, off_t start);
    void clear();
    int open_mirror_file();
    bool set_all_status(bool is_loaded);
    bool set_all_status_unloaded() {
        return set_all_status(false);
    }

private:
    pthread_mutex_t     _entity_lock;
    std::string         _path;             // remote object path
    std::atomic<int>    _fd;               // cache file fd
    std::atomic<int>    _ref_count;
    std::atomic<bool>   _is_modified;      // status for file changed
    ObjectMetaData      _origin_meta;      // original meta headers
    size_t              _origin_meta_size; // original file size in original headers
    bool                _is_tmpfile;       // local tmp file, will be removed on close
    std::string         _tmp_filename;     // temp file name
    ERwStrategy         _rw_strategy;      // logic read/write strategy


    pthread_mutex_t     _entity_data_lock;
    bool                _use_mem_cache;    // use memory cache model
    MemBuf              _mem_buf;          // memory buffer
    MultiUploadHelper*  _multiupload_helper;

    ObjectPageList      _page_list;        // file data page
    std::string         _cache_path;       // local cache file path
    std::string         _mirror_path;      // mirror file path to local cache file
};


class DataCache {
public:
    typedef std::map<std::string, DataCacheEntity *> DataCacheMap;

    static DataCache *instance() {
        return &_s_instance;
    }
    static size_t get_ensure_free_disk_space() {
        return _free_disk_space;
    }
    static size_t set_ensure_free_disk_space(size_t size);
    static size_t init_ensure_free_disk_space() {
        return set_ensure_free_disk_space(0);
    }

private:
    static DataCache _s_instance;
    static size_t _free_disk_space;

public:
    DataCache();
    ~DataCache();

    inline bool is_cache_dir() const {
        return !_cache_dir.empty();
    }
    const char *get_cache_dir() const {
        return _cache_dir.c_str();
    }
    void set_tmp_dir(const std::string &dir) {
        _tmp_dir = dir;
    }
    const std::string &tmp_dir() const {
        return _tmp_dir;
    }
    const char *tmp_dir_cstr() const {
        return _tmp_dir.c_str();
    }

    int set_cache_dir(const std::string &dir);
    bool delete_cache_dir();
    int delete_cache_file(const char *path);
    bool make_cache_path(
            const char *path,
            std::string &cache_path,
            bool is_create_dir = true,
            bool is_mirror_path = false);
    bool check_cache_top_dir();

    DataCacheEntity *get_cache(const char *path);

    DataCacheEntity *open_cache(
            const char *path,
            ObjectMetaData *pmeta = NULL,
            ssize_t size = -1,
            time_t time = -1,
            bool force_tmpfile = false,
            bool is_create = true);

    DataCacheEntity *exist_open(const char *path);

    bool close_cache(DataCacheEntity *ent);

private:
    pthread_mutex_t   _data_cache_lock;
    DataCacheMap      _data_cache;
    std::string       _cache_dir;
    std::string       _tmp_dir;
};

END_FS_NAMESPACE

#endif
/* vim: set expandtab ts=4 sw=4 sts=4 tw=100: */
