/***************************************************************************
 *
 * Copyright (c) 2017 Baidu.com, Inc. All Rights Reserved
 *
 **************************************************************************/
/**
 * @file src/file_manager.cpp
 * @author lili19(com@baidu.com)
 * @date 2017/09/12 09:35:27
 * @brief
 *
 **/
#include "file_manager.h"
#include "bosfs_util.h"
#include <algorithm>

BEGIN_FS_NAMESPACE

int File::load_meta_from_bos() {
    MutexGuard lock(&_mutex);
    int ret = BosfsUtil::head_object(_name.substr(1), &_meta, &_is_dir_obj, &_is_prefix);
    if (ret == 0) {
        _load_time_us = get_system_time_us();
    }
    return ret;
}

// compatible with old user meta which not has "bosfs-" prefix
int File::stat(struct stat *st) {
    MutexGuard lock(&_mutex);
    BosfsUtil::init_default_stat(st);

    // is dir
    bool is_dir = false;
    std::string content_type = _meta.content_type();
    if (_is_dir_obj && _meta.content_length() == 0) {
        is_dir = true;
    } else if (!content_type.empty()) {
        content_type = content_type.substr(0, content_type.find(';'));
        is_dir = (content_type == "application/x-directory");
    }

    if (_name == "/" || (!is_dir && _is_prefix)) {
        st->st_size = ST_BLKSIZE;
        st->st_blocks = ST_MINBLOCKS;
        return 0;
    }

    //size
    st->st_size = _meta.content_length();
    st->st_blocks = (st->st_size + st->st_blksize - 1) / st->st_blksize * ST_MINBLOCKS;

    //mtime
    const std::string *meta_mtime = &_meta.user_meta("bosfs-mtime");
    if (meta_mtime->empty()) {
        meta_mtime = &_meta.user_meta("mtime");
    }
    if (meta_mtime->empty()) {
        st->st_mtime = _meta.last_modified();
    } else {
        st->st_mtime = strtol(meta_mtime->c_str(), NULL, 0);
    }
    st->st_ctime = st->st_mtime;
    st->st_atime = st->st_mtime;

    //uid gid: use default
    // const std::string *meta_uid = &_meta.user_meta("bosfs-uid");
    // if (meta_uid->empty()) {
    //     meta_uid = &_meta.user_meta("uid");
    // }
    // if (!meta_uid->empty()) {
    //     st->st_uid = strtol(meta_uid->c_str(), NULL, 0);
    // }
    // const std::string *meta_gid = &_meta.user_meta("bosfs-gid");
    // if (meta_gid->empty()) {
    //     meta_gid = &_meta.user_meta("gid");
    //     use default
    // }
    // if (!meta_gid->empty()) {
    //     st->st_gid = strtol(meta_gid->c_str(), NULL, 0);
    // }

    // st_mode 
    st->st_mode = g_mount_mode;
    const std::string *meta_mode = &_meta.user_meta("bosfs-mode");
    if (meta_mode->empty()) {
        meta_mode = &_meta.user_meta("mode");
    }
    if (!meta_mode->empty()) {
        st->st_mode = strtol(meta_mode->c_str(), NULL, 0);
        if (!(st->st_mode & S_IFMT)) {//前四位表示文件类型
            st->st_mode |= is_dir ? S_IFDIR : S_IFREG;
        }
    } else {
        if (!is_dir) {
            st->st_mode &= ~(S_IFMT | 0111);//低三位
            st->st_mode |= S_IFREG;
        } else {
            st->st_size = ST_BLKSIZE;
            st->st_blocks = ST_MINBLOCKS;
        }
    }
    return 0;
}

FileManager FileManager::_s_instance;

int FileManager::get(const std::string &name, FilePtr *file) {
    if (try_get(name, file)) {
        return 0;
    }
    file->reset(new File(name));
    int ret = (*file)->load_meta_from_bos();
    if (ret != 0) {
        if (ret == BOSFS_OBJECT_NOT_EXIST) {
            return -ENOENT;
        }
        return -EIO;
    }
    size_t size = 0;
    {
        bcesdk_ns::TLSLockWriteGuard lock(_lock);
        FileTable::iterator it = _table.find(name);
        if (it != _table.end()) {
            *file = it->second;
        } else {
            _table[name] = *file;
        }
        size = _table.size();
    }
    //BOSFS_DEBUG("[meta add] [name=%s] add new meta.", name.c_str());
    if (_cache_capacity >= 0 && size > (size_t) _cache_capacity) {
        gc();
    }
    return 0;
}

bool FileManager::try_get(const std::string &name, FilePtr *file) { 
    bcesdk_ns::TLSLockWriteGuard lock(_lock);
    FileTable::iterator it = _table.find(name);
    if (it == _table.end()) {
        return false;
    }
    int64_t now = get_system_time_us();
    //BOSFS_DEBUG("[meta try_get] [name=%s] _expire_us:%ld, load_time:%ld, now:%ld", it->second->name().c_str(), _expire_us, it->second->load_time_us(), now);
    if (_expire_us >= 0 && (it->second->load_time_us() + _expire_us) < now) {
        if (it->second.use_count() <= 1) {
            _table.erase(it);
            return false;
        }
    }
    it->second->hit(now);
    *file = it->second; 
    return true;
}

// touch by readdir(list dri/file), create new file
void FileManager::set(const std::string &name, FilePtr &file, bool is_sync_gc) {
    size_t size = 0;
    {
        bcesdk_ns::TLSLockWriteGuard lock(_lock);
        _table[name] = file;
        size = _table.size();
    }
    if (is_sync_gc && _cache_capacity >= 0 && size > (size_t) _cache_capacity) {
        gc();
    }
}

void FileManager::del(const std::string &name) {
    bcesdk_ns::TLSLockWriteGuard lock(_lock);
    BOSFS_DEBUG("[meta del] [name=%s] del meta", name.c_str());
    _table.erase(name);
}


void FileManager::try_gc() { 
    size_t size = 0;
    {
        bcesdk_ns::TLSLockReadGuard lock(_lock);
        size = _table.size();

    }
    if (_cache_capacity >= 0 && size > (size_t) _cache_capacity) {
        BOSFS_DEBUG("[meta try gc] [size:%d]", size);
        gc();
    }
}


void FileManager::gc() {
    int64_t now = get_system_time_us();
    //std::vector<FilePtr> candis;
    std::map<int64_t, FilePtr> candis;
    std::vector<FilePtr> removes;
    size_t asize = 0;

    {
        bcesdk_ns::TLSLockReadGuard lock(_lock);
        for (FileTable::iterator it = _table.begin(); it != _table.end(); ++it) {
            if (it->second) {
                if (_expire_us >= 0 && (it->second->load_time_us() + _expire_us) < now) {
                    removes.push_back(it->second);
                } else {
                    //candis.push_back(it->second);
                    candis.insert({it->second->hit_time_s(), it->second});
                }
            } else {
                BOSFS_ERR("meta data is NULL!");
            }
        }
        asize = _table.size();
    }

    if (_cache_capacity >= 0 && candis.size() > (size_t) _cache_capacity) {
        auto it = candis.begin();
        //std::sort(candis.begin(), candis.end(), compare_cache_priority);
        for (int i = _cache_capacity; it != candis.end() && i < candis.size(); i++, it++) {
            if (it->second) {
                removes.push_back(it->second);
            }
        }
    }
    //BOSFS_DEBUG("[meta gcc] candis size:%d, removes size:%d, meta all size:%d",candis.size(), removes.size(), asize);

    candis.clear();

    {
        bcesdk_ns::TLSLockWriteGuard lock(_lock);
        for (size_t i = 0; i < removes.size(); ++i) {
            FileTable::iterator it = _table.find(removes[i]->name());
            if (it == _table.end()) {
                continue;
            }
            if (it->second.use_count() > 2) {
                continue;
            }
            //BOSFS_DEBUG("[meta gc] [name=%s] do gc meta.", it->first.c_str());
            _table.erase(it);
        }
    }
}

END_FS_NAMESPACE

