/**
 * Copyright 2014 (c) Baidu, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>

#include <string>
#include <iostream>
#include <stdarg.h>
#include <signal.h>

#include "bosfs.h"
#include "file_manager.h"

#ifndef PACKAGE_STRING
#define PACKAGE_STRING "bosfs " VERSION
#endif
#define LICENSE "Copyright (c) 2014 Baidu.com, Inc. All rights reserved."

BEGIN_FS_NAMESPACE

struct BosfsConfItem {
    BosfsConfItem() : is_set(false) {}
    BosfsConfItem(const std::string &short_name,
            const std::string &value_prompt) : is_set(false) {
        this->short_name = short_name;
        this->value_prompt = value_prompt;
    }
    BosfsConfItem(const std::string &short_name,
            const std::string &value_prompt,
            const std::string &description) : is_set(false) {
        this->short_name = short_name;
        this->value_prompt = value_prompt;
        this->description = description;
    }
    std::string short_name;

    bool is_set;
    std::string value;
    std::string value_prompt;
    std::string description;
};

enum FuseArgsKey {
    FUSE_O_UID,
    FUSE_O_GID,
    FUSE_O_UMASK,
    FUSE_O_ALLOW_OTHER,
    FUSE_O_DIRECTIO
};

// fuse arguments that bos cares
std::map<std::string, int> g_fuse_args;
// bos own arguments
std::map<std::string, BosfsConfItem> g_bos_args;
// short name map
std::map<std::string, std::string> g_old_bos_args;

void assert_local_dir(const char *name, const std::string &localdir) {//防止本地的内容刷入bos，如logfile和临时文件
    struct stat st;
    if (stat(localdir.c_str(), &st) != 0) {
        die("unable to access local directory of %s:%s, error:%s", name, localdir.c_str(),
                strerror(errno));
    }
    if (!(S_ISDIR(st.st_mode))) {
        die("local directory of %s:%s is not a directory", name, localdir.c_str());
    }
    char buffer[10240];
    if (realpath(localdir.c_str(), buffer) == NULL) {
        die("unable get absolute path of local directory %s:%s, errno:%d", name, localdir.c_str(),
                errno);
    }
    std::string dir = buffer;
    std::string parent = g_mountpoint + "/";
    if (dir.compare(0, parent.length(), parent) == 0) {
        die("local directory of %s:%s can't inside mountpoint:%s", name, localdir.c_str(),
                parent.c_str());
    }
}

void init_bos_args() {
    g_fuse_args["uid"] = FUSE_O_UID;
    g_fuse_args["gid"] = FUSE_O_GID;
    g_fuse_args["umask"] = FUSE_O_UMASK;
    g_fuse_args["allow_other"] = FUSE_O_ALLOW_OTHER;
    g_fuse_args["direct_io"] = FUSE_O_DIRECTIO;

    g_bos_args["bos.fs.log.file"] = BosfsConfItem("logfile",
            "file path like xxx/xxx.log");
    g_bos_args["bos.fs.log.level"] = BosfsConfItem("loglevel",
            "level: fatal,error,warn,info,debug; case ignored");
    g_bos_args["bos.fs.log.rotate_size"] = BosfsConfItem("log_rotate_size",
            "log rotate size, use unit MB",
            "default is 1024MB");
    g_bos_args["bos.fs.endpoint"] = BosfsConfItem("endpoint",
            "url like http://bj.bcebos.com, http:// can be omitted",
            "specify server address, use https:// for SSL. if not specified, bucket's location would be used; "
            "default is bj.bcebos.com");
    g_bos_args["bos.fs.ak"] = BosfsConfItem("ak",
            "your ak");
    g_bos_args["bos.fs.sk"] = BosfsConfItem("sk",
            "your sk");
    g_bos_args["bos.fs.credentials"] = BosfsConfItem("credentials",
            "your credential file path");
    g_bos_args["bos.fs.multipart_parallel"] = BosfsConfItem("multipart_parallel", "limit the client maximum multipart parallel requests send to the server, default is 10");
    g_bos_args["bos.fs.cache.base"] = BosfsConfItem("use_cache",
            "cache directory in absolute path");
    g_bos_args["bos.fs.meta.expires"] = BosfsConfItem("meta_expires",
            "seconds", "after how many seconds the local meta will be expired, default is infinite");
    g_bos_args["bos.fs.meta.capacity"] = BosfsConfItem("meta_capacity",
            "integer number", "how many meta cache items will be keeped as a hit, default is 100000");
    g_bos_args["bos.fs.meta.strategy"] = BosfsConfItem("meta_strategy", "strategy: remote,merge",
            "meta strategy, default is merge");
    g_bos_args["bos.fs.storage_class"] = BosfsConfItem("storage_class",
            "standard or standard_ia; case ignored",
            "when specified this option, any upload action will use the storage class");
    g_bos_args["bos.fs.createbucket"] = BosfsConfItem("createbucket", "",
            "create bucket if not exist when mounting");
    g_bos_args["bos.fs.tmpdir"] = BosfsConfItem("tmpdir", "an existing directory in absolute path",
            "specified where bosfs creates temporary file in, default is /tmp");
    g_bos_args["bos.sdk.multipart_size"] = BosfsConfItem("multipart_size", "number small than 5GB, can use unit KB,MB",
            "an hint to part size in multiple upload, default is 10MB");
    g_bos_args["bos.sdk.multipart_threshold"] = BosfsConfItem("multipart_threshold", "number small than 5GB, can use unit KB,MB",
            "when file is larger than this value, multiple upload will be used, default is 100MB");
    g_bos_args["bos.fs.preload_blocks"] = BosfsConfItem("preload_blocks",
            "1MB/block", "read preload blocks cnt, preload data size = preload_blocks * 1MB, default is 0");
    g_bos_args["bos.fs.connect.timeout"] = BosfsConfItem("connect_timeout",
            "milliseconds", "time limit for connect, default is 1000");
    g_bos_args["bos.fs.client.timeout"] = BosfsConfItem("client_timeout",
            "milliseconds", "time limit for request, default is 1200000");
    g_bos_args["bos.fs.mem_cache_per_size"] = BosfsConfItem("mem_cache_per_size", "MB",
            "switch to use memory cache model when is not default 0. it's the buffer size in One open file, default is 0(means not to use mem cache model).");
    g_bos_args["bos.fs.mem_cache_limit_size"] = BosfsConfItem("mem_cache_limit_size", "MB",
            "all memory limit used by the bosfs process in memory cache buffer(=mem_cache_per_size * opening file numbers), default is 80% system free memory size.");
    g_bos_args["bos.fs.rw_strategy"] = BosfsConfItem("rw_strategy", "read/write strategy: mem_random,mem_sequence,preload_write",
            "must set mem_cache_per_size when using mem_random/mem_sequence! [mem_random]: random read or write in small file that file size <= mem_cache_per_size"
            " [mem_sequence]: in order to read or write in super file that means download/upload scenario"
            " [preload_write]: load all file from bos before write");

    for (std::map<std::string, BosfsConfItem>::iterator it = g_bos_args.begin();
            it != g_bos_args.end(); ++it) {
        if (!it->second.short_name.empty()) {
            g_old_bos_args[it->second.short_name] = it->first;
        }
    }
}

void parse_bos_args() {
    std::string name;
    if (g_bos_args["bos.fs.log.file"].is_set) {
        std::string logfile = g_bos_args["bos.fs.log.file"].value;
        std::string logdir = ".";
        size_t pos = logfile.rfind('/');
        if (pos != std::string::npos) {
            logdir = logfile.substr(0, pos);
        }
        assert_local_dir("log", logdir);
        BosfsUtil::set_log_file(logfile);
    }
    if (g_bos_args["bos.fs.log.level"].is_set) {
        BosfsUtil::set_log_level(g_bos_args["bos.fs.log.level"].value);
    }
    if (g_bos_args["bos.fs.log.rotate_size"].is_set) {
        int rotate_size = 0;
        if (!StringUtil::str2int(g_bos_args["bos.fs.log.rotate_size"].value, &rotate_size)) {
            die("invalid log rotate size: %s", g_bos_args["bos.fs.log.rotate_size"].value.c_str());
            return;
        }
        if (rotate_size <= 0) {
            die("invalid log rotate size: %d", rotate_size);
            return;
        }
        rotate_size = rotate_size * 1024 * 1024;
        BosfsUtil::set_log_rotate_size(rotate_size);
    }
    if (g_bos_args["bos.fs.endpoint"].is_set) {
        g_endpoint = g_bos_args["bos.fs.endpoint"].value;
    }
    if (g_bos_args["bos.fs.credentials"].is_set) {
        g_credentials_path = g_bos_args["bos.fs.credentials"].value;
    } else {
        char* home= getenv("HOME");
        if (home) {
            g_credentials_path = std::string(home) + "/.bce/credentials";
        }
    }
    if (g_bos_args["bos.fs.ak"].is_set) {
        g_ak = g_bos_args["bos.fs.ak"].value;
    }
    if (g_bos_args["bos.fs.sk"].is_set) {
        g_sk = g_bos_args["bos.fs.sk"].value;
    }
    if (g_bos_args["bos.fs.cache.base"].is_set) {
        std::string cache_dir = g_bos_args["bos.fs.cache.base"].value;
        assert_local_dir("log", cache_dir);
        int ret = DataCache::instance()->set_cache_dir(cache_dir);
        if (ret != 0) {
            die("set cache dir %s failed: %d", cache_dir.c_str(), ret);
            return;
        }
    }
	if (g_bos_args["bos.fs.meta.expires"].is_set) {
		int secs = 0;
        int64_t us = 0;
		if (!StringUtil::str2int(g_bos_args["bos.fs.meta.expires"].value, &secs)) {
            die("invalid number: %s", g_bos_args["bos.fs.meta.expires"].value.c_str());
			return;
		}
        if (secs == 0) {
            us = g_meta_expire_delay_us;
        } else if (secs == -1) {
            us = 0; 
        } else {
            us = secs * 1000000;
        }
        FileManager::instance().set_expire_us(us);
    } else {
        // default value: g_meta_expire_delay_us
        // FileManager::instance().set_expire_us(g_meta_expire_delay_us);
    }
    if (g_bos_args["bos.fs.meta.capacity"].is_set) {
		int num;
		if (!StringUtil::str2int(g_bos_args["bos.fs.meta.capacity"].value, &num)) {
            die("invalid number: %s", g_bos_args["bos.fs.meta.capacity"].value.c_str());
			return;
		}
        FileManager::instance().set_cache_capacity(num);
    } else {
        FileManager::instance().set_cache_capacity(100000);
    }
    if (g_bos_args["bos.fs.meta.strategy"].is_set) {
        std::string strategy = StringUtil::upper(g_bos_args["bos.fs.meta.strategy"].value);
        if (strategy == "LOCAL") {
            g_meta_strategy = EMetaStrategy::META_LOCAL;
        } else if (strategy == "REMOTE") {
            g_meta_strategy = EMetaStrategy::META_REMOTE;
        } else if (strategy == "MERGE") {
            g_meta_strategy = EMetaStrategy::META_MERGE;
        } else if (strategy == "") {
            g_meta_strategy = EMetaStrategy::META_MERGE;
        } else {
            die("invalid meta strategy: %s", g_bos_args["bos.fs.meta.strategy"].value.c_str());
            return;
        }
    } else {
        g_meta_strategy = EMetaStrategy::META_MERGE;
    }
    if (g_bos_args["bos.fs.createbucket"].is_set) {
        g_create_bucket = true;
    }

    if (g_bos_args["bos.fs.storage_class"].is_set) {
        g_storage_class = StringUtil::upper(g_bos_args["bos.fs.storage_class"].value);
        if (g_storage_class != "STANDARD" && g_storage_class != "STANDARD_IA") {
            die("invalid storage class: %s", g_bos_args["bos.fs.storage_class"].value.c_str());
            return;
        }
    }
	if (g_bos_args["bos.fs.multipart_parallel"].is_set) {
        if (!StringUtil::str2int(g_bos_args["bos.fs.multipart_parallel"].value, &g_multipart_parallel)) {
            die("%s: invalid number string %s", "bos.fs.multipart_parallel", g_bos_args["bos.fs.multipart_parallel"].value.c_str());
            return;
        }
    }
    if (g_bos_args["bos.fs.tmpdir"].is_set) {
        const std::string &dir = g_bos_args["bos.fs.tmpdir"].value;
        assert_local_dir("tmp", dir);
        DataCache::instance()->set_tmp_dir(dir);
    } else {
        assert_local_dir("default tmp", "/tmp");
        DataCache::instance()->set_tmp_dir("/tmp");
    }
    name = "bos.sdk.multipart_size";
    if (g_bos_args[name].is_set) {
        if (!StringUtil::byteunit2int(g_bos_args[name].value, &g_multipart_size)) {
            die("%s: invalid number string:%s", name.c_str(), g_bos_args[name].value.c_str());
            return;
        }
    }
    name = "bos.sdk.multipart_threshold";
    if (g_bos_args[name].is_set) {
        if (!StringUtil::byteunit2int(g_bos_args[name].value, &g_multipart_threshold)) {
            die("%s: invalid number string:%s", name.c_str(), g_bos_args[name].value.c_str());
            return;
        }
    }
    if (g_bos_args["bos.fs.block_size"].is_set) {
        if (!StringUtil::byteunit2int(g_bos_args["bos.fs.block_size"].value, &g_block_size)) {
            die("%s: invalid number string:%s", "bos.fs.block_size", g_bos_args["bos.fs.block_size"].value.c_str());
            return;
        }
    }
    if (g_bos_args["bos.fs.preload_blocks"].is_set) {
        if (!StringUtil::byteunit2int(g_bos_args["bos.fs.preload_blocks"].value, &g_preload_blocks)) {
            die("%s: invalid number string:%s", "bos.fs.preload_blocks", g_bos_args["bos.fs.preload_blocks"].value.c_str());
            return;
        }
    }
    if (g_bos_args["bos.fs.connect.timeout"].is_set) {
        if (!StringUtil::byteunit2int(g_bos_args["bos.fs.connect.timeout"].value, &g_connect_timeout)) {
            die("%s: invalid number string:%s", "bos.fs.connect.timeout", g_bos_args["bos.fs.connect.timeout"].value.c_str());
            return;
        }
    }
    if (g_bos_args["bos.fs.client.timeout"].is_set) {
        if (!StringUtil::byteunit2int(g_bos_args["bos.fs.client.timeout"].value, &g_client_timeout)) {
            die("%s: invalid number string:%s", "bos.fs.client.timeout", g_bos_args["bos.fs.client.timeout"].value.c_str());
            return;
        }
    }
    if (g_bos_args["bos.fs.mem_cache_per_size"].is_set) {
        if (!StringUtil::byteunit2int(g_bos_args["bos.fs.mem_cache_per_size"].value, &g_mem_cache_per_size)) {
            die("%s: invalid number string:%s", "bos.fs.mem_cache_per_size", g_bos_args["bos.fs.mem_cache_per_size"].value.c_str());
            return;
        }
        if (g_mem_cache_per_size < g_preload_blocks) {
            die("bos.fs.mem_cache_per_size: invalid size:%s, must >= bos.fs.preload_blocks[%ld]", g_bos_args["bos.fs.mem_cache_per_size"].value.c_str(), g_preload_blocks);
            return;
        }
        g_mem_cache_per_size *= (1024 * 1024); //B
        if (g_mem_cache_per_size < g_multipart_min_threshold) {
            die("bos.fs.mem_cache_per_size: invalid size:%s, must >= 5MB", "bos.fs.mem_cache_per_size", g_bos_args["bos.fs.mem_cache_per_size"].value.c_str());
            return;
        }
    }

    if (g_bos_args["bos.fs.mem_cache_limit_size"].is_set) {
        if (!StringUtil::byteunit2int(g_bos_args["bos.fs.mem_cache_limit_size"].value, &g_mem_cache_size_limit)) {
            die("%s: invalid number string:%s", "bos.fs.mem_cache_limit_size", g_bos_args["bos.fs.mem_cache_limit_size"].value.c_str());
            return;
        }
        g_mem_cache_size_limit *= (1024 * 1024); //B
    } else {
        g_mem_cache_size_limit = get_free_memory() /*kB*/ * 1024 /*B*/ * 80 / 100 /*80%*/;
    }

    if (g_bos_args["bos.fs.rw_strategy"].is_set) {
        std::string strategy = StringUtil::upper(g_bos_args["bos.fs.rw_strategy"].value);
        if (strategy == "MEM_RANDOM") {
            g_rw_strategy = ERwStrategy::S_RANDOM_RW;
        } else if (strategy == "MEM_SEQUENCE") {
            g_rw_strategy = ERwStrategy::S_MEM_SEQUENCE_RW;
        }  else if (strategy == "PRELOAD_WRITE") {
            g_rw_strategy = ERwStrategy::S_PRELOAD_BEFORE_W;
        } else {
            die("invalid mem cache model strategy: %s", g_bos_args["bos.fs.rw_strategy"].value.c_str());
            return;
        }
    }

    // check if use mem cache model
    if (g_mem_cache_per_size) {
        if (g_rw_strategy == ERwStrategy::S_DEFAUL) {
            die("must use '-o rw_strategy=mem_random/mem_secquence' rw strategy when mem_cache_limit_size is setting!");
        }
        if (!g_is_direct_io) {
            die("must use '-o direct_io' when mem_cache_limit_size is setting!");
        }
    }
}

void show_version() {
    printf("%s\n", PACKAGE_STRING);
#ifdef BOSFS_REPO
    printf("repo: %s\n", BOSFS_REPO);
#endif
    printf("build time: %s %s\n", __DATE__, __TIME__);
    printf("%s\n", LICENSE);
    exit(0);
}
void show_help() {
    std::cout << "Usage: " << g_program_name << " bucket mountpoint [OPTIONS]" << std::endl
            << "OPTIONS can be fuse options, or bos options below:" << std::endl;
    std::cout << "BOS options:" << std::endl;

    for (std::map<std::string, BosfsConfItem>::iterator it = g_bos_args.begin();
            it != g_bos_args.end(); ++it) {
        std::cout << "\t";
        if (!it->second.short_name.empty()) {
            std::cout << "-o " << it->second.short_name << ", ";
        }
        std::cout << "-o " << it->first;
        if (!it->second.value_prompt.empty()) {
            std::cout << "=<" << it->second.value_prompt << ">";
        }
        std::cout << std::endl;
        if (!it->second.description.empty()) {
            std::cout << "\t\t" << it->second.description << std::endl;
        }
    }
    fprintf(stdout, "FUSE options:\n"
            "\t-f foreground mode\n"
            "\t-d debug mode\n"
            "\t-o ro\tread only mode\n"
            "\t-o direct_iouse direct_io mode\n"
            "\t-o fsname=<filesystem name>, shows in df command\n"
            "\t-o allow_other allow other user access mountpoint\n"
            "\t-o mount_umask when uses allow_other, use this to forbid permissions, default is 022\n");
    exit(0);
}


//命令参考：bosfs mybucket my_local_directory -o endpoint=http://bj.bcebos.com
//-o ak=xxxxxxxxxxxxxxxx -o sk=xxxxxxxxxxxxxxxxx -o logfile=xx/xx.log
int g_noopt_arg_index = 0;
int fuse_opt_handler(void * /*data*/, const char *arg, int key, struct fuse_args *outargs) {
    (void) outargs;
    std::string argstr(arg);
    switch (key) {
    //不带-的参数，解析bucket和挂载点
    case FUSE_OPT_KEY_NONOPT: {
        switch (g_noopt_arg_index++) {
        case 0: {//第一个参数为bucket name 或者子目录
            const std::string bucket_path(arg);
            std::size_t pos = bucket_path.find("/");
            g_bucket = bucket_path.substr(0, pos);
            if (pos != std::string::npos) {
                g_bucket_prefix = bucket_path.substr(pos+1);
                if (*g_bucket_prefix.rbegin() != '/' && g_bucket_prefix != "") {
                    g_bucket_prefix.push_back('/');
                }
            }
            return 0;
         }
        case 1: {//挂载点
            char buffer[10240];
            if (realpath(arg, buffer) == NULL) {
                die("unable get absolute path of mountpoint:%s, errno:%d", arg, errno);
            }
            g_mountpoint = buffer;
            return 1;
        }
        default:
            return 1;
        }
        break;
    }
    case FUSE_OPT_KEY_OPT: {
        std::string key;
        std::string value;
        size_t pos = argstr.find('=');
        key = StringUtil::trim(argstr.substr(0, pos));
        if (pos != std::string::npos) {//-o ak=xxxx -o sk=xxx etc..
            value = StringUtil::trim(argstr.substr(pos + 1));
        }
        if (g_old_bos_args.find(key) != g_old_bos_args.end()) {
            key = g_old_bos_args[key];
        }
        if (g_bos_args.find(key) != g_bos_args.end()) {
            g_bos_args[key].is_set = true;
            g_bos_args[key].value = value;
            return 0;
        }
        if (g_fuse_args.find(key) != g_fuse_args.end()) {
            int fuse_key = g_fuse_args[key];
            switch (fuse_key) {
            case FUSE_O_UID:
                g_bosfs_uid = strtol(value.c_str(), NULL, 0);
                if (geteuid() != 0) { //root uid=0
                    die("only root user can specify uid");
                }
                g_is_bosfs_uid = true;
                break;
            case FUSE_O_GID:
                g_bosfs_gid = strtol(value.c_str(), NULL, 0);
                if (geteuid() != 0) {
                    die("only root user can specify gid");
                }
                g_is_bosfs_gid = true;
                break;
            case FUSE_O_UMASK:
                g_bosfs_mask = strtol(value.c_str(), NULL, 0);
                g_bosfs_mask &= (S_IRWXU | S_IRWXG | S_IRWXO);
                g_is_bosfs_umask = true;
                break;
            case FUSE_O_ALLOW_OTHER:
                g_allow_other = true;
                break;
            case FUSE_O_DIRECTIO:
                g_is_direct_io = true;
                break;
            }
            return 1;
        }
        if (key == "mount_umask") {
            g_mount_umask = strtol(value.c_str(), NULL, 0);
            g_mount_umask &= (S_IRWXU | S_IRWXG | S_IRWXO);
            g_is_mount_umask = true;
            return 0;
        }
        if (argstr == "-h" || argstr == "--help") {
            show_help();
        }
        if (argstr == "-v" || argstr == "--version") {
            show_version();
        }
        break;
    }
    }
    return 1;
}

int bosfs_main(int argc, char *argv[]) {
    g_program_name = argv[0];
    init_bos_args();

    // Pass custom args to fuse args handler
    struct fuse_args custom_args = FUSE_ARGS_INIT(argc, argv);
    if (0 != fuse_opt_parse(&custom_args, NULL, NULL, fuse_opt_handler)) {
        exit(EXIT_FAILURE);
    }

    // Start to check custom args after fuse_opt_handler parsing
    if (g_bucket.size() == 0) {
        die("missing BUCKET argument");
    }
    if (g_bucket.find_first_of("/:\\;!@#$%^&*?") != std::string::npos) {
        die("bucket name(%s) contains illegal letter", g_bucket.c_str());
    }

    struct stat stbuf;
    if (-1 == stat(g_mountpoint.c_str(), &stbuf)) {
        BOSFS_FATAL("unable to access MOUNTPOINT %s: %s", g_mountpoint.c_str(), strerror(errno));
        return -1;
    }
    if (!(S_ISDIR(stbuf.st_mode))) {
        BOSFS_FATAL("MOUNTPOINT: %s is not a directory", g_mountpoint.c_str());
        return -1;
    }
    parse_bos_args();

    if (!BosfsUtil::set_mountpoint_attr(stbuf)) {
        BOSFS_FATAL("MOUNTPOINT: %s permission denied", g_mountpoint.c_str());
        return -1;
    }
    if (!DataCache::instance()->check_cache_top_dir() || !StatCacheFile::check_top_dir()) {
        die("can't check permission of cache directory");
    }
    DataCache::init_ensure_free_disk_space();

    // Delete cache dirs
    if (g_remove_cache) {
        if (!StatCacheFile::delete_dir()) {
            BOSFS_WARN("%s", "could not initialize cache directory");
        }
        if (!DataCache::instance()->delete_cache_dir()) {
            BOSFS_WARN("%s", "could not initialize cache directory");
        }
    }

    // Initiate mime-types
    if (!BosfsUtil::init_mimetype()) {
        BOSFS_ERR("%s", "could not initiate mime-types");
        return 2;
    }

    // move to fuse init callback
    // if (BosfsUtil::init_bos() != 0) {
    //     return 3;
    // }
    signal(SIGUSR2, BosfsUtil::reload_sighandler);

    // Link the file system interface to fuse
    struct fuse_operations bosfs_operation;
    memset(&bosfs_operation, 0, sizeof(struct fuse_operations));
    bosfs_operation.init        = Bosfs::init;
    bosfs_operation.destroy     = Bosfs::destroy;
    bosfs_operation.access      = Bosfs::access;
    bosfs_operation.create      = Bosfs::create;
    bosfs_operation.open        = Bosfs::open;
    bosfs_operation.read        = Bosfs::read;
    bosfs_operation.write       = Bosfs::write;
    bosfs_operation.statfs      = Bosfs::statfs;
    bosfs_operation.flush       = Bosfs::flush;
    bosfs_operation.fsync       = Bosfs::fsync;
    bosfs_operation.release     = Bosfs::release;
    bosfs_operation.symlink     = Bosfs::symlink;
    bosfs_operation.link        = Bosfs::link;
    bosfs_operation.unlink      = Bosfs::unlink;
    bosfs_operation.readlink    = Bosfs::readlink;
    bosfs_operation.mknod       = Bosfs::mknod;
    bosfs_operation.mkdir       = Bosfs::mkdir;
    bosfs_operation.rmdir       = Bosfs::rmdir;
    bosfs_operation.rename      = Bosfs::rename;
    bosfs_operation.opendir     = Bosfs::opendir;
    bosfs_operation.readdir     = Bosfs::readdir;
    bosfs_operation.releasedir  = Bosfs::releasedir;
    bosfs_operation.chmod       = Bosfs::chmod;
    bosfs_operation.chown       = Bosfs::chown;
    bosfs_operation.utimens     = Bosfs::utimens;
    bosfs_operation.truncate    = Bosfs::truncate;
    bosfs_operation.getattr     = Bosfs::getattr;
    bosfs_operation.listxattr   = Bosfs::listxattr;
    bosfs_operation.getxattr    = Bosfs::getxattr;
    bosfs_operation.setxattr    = Bosfs::setxattr;
    bosfs_operation.removexattr = Bosfs::removexattr;
    int ret = fuse_main(custom_args.argc, custom_args.argv, &bosfs_operation, NULL);
    fuse_opt_free_args(&custom_args);
    BosfsUtil::stop_reload();
    return ret;
}

END_FS_NAMESPACE

int main(int argc, char *argv[]) {
    return bosfs_ns::bosfs_main(argc, argv);
}
