/***************************************************************************
 *
 * Copyright (c) 2021 Baidu.com, Inc. All Rights Reserved
 *
 **************************************************************************/
/**
 * @file src/mem_buf.cpp
 * @author caixiayu@baidu.com
 * @date   2021.07
 **/

#include <string.h>
#include "mem_buf.h"

BEGIN_FS_NAMESPACE

MemBuf::MemBuf() : _buffer(nullptr), _size(0), _pos(0), _file_size(0), _data_pos(0), _pre_load_pos(0),
                   _last_op(EMemCacheOp::OP_DEFAULT), _last_op_size(0), _last_op_pos(0),
                   _rw_strategy(ERwStrategy::S_DEFAUL) {
}

MemBuf::~MemBuf() {
    if (_buffer) {
        delete[] _buffer;
        _buffer = nullptr;
    }
    g_mem_cache_size_use -= _size;
}

void MemBuf::init(int64_t buf_size, int64_t file_size, int64_t pos) {
    _size = buf_size;
    _file_size = file_size;
    _pos = pos;
    _buffer = new char[_size];
    g_mem_cache_size_use += _size;
}

int64_t MemBuf::read(char *buffer, int64_t size, int64_t offset) {
    if (offset > _file_size) {
        return 0;
    }
    if (_rw_strategy == ERwStrategy::S_RANDOM_RW) {
        _pos = offset;
        size = remain_size(size);
    } else if (_rw_strategy == ERwStrategy::S_MEM_SEQUENCE_RW) {
        if (!check_sequence(EMemCacheOp::OP_READ, offset)) {
            BOSFS_ERR("read is not sequential. last op:%ld offset:%ld, size:%ld, cur offset:%ld",
             _last_op, _last_op_pos, _last_op_size, offset);
             return -1;
        }
        if (offset + size > _file_size) {
            size = _file_size - offset;
            BOSFS_DEBUG("cut read size to:%ld, offset:%ld, file_size:%ld", size, offset, _file_size);
        }
        size = remain_size(size);
        update_last_op(EMemCacheOp::OP_READ, offset, size);
    }
    memcpy(buffer, (const char *)_buffer + _pos, (size_t)size);
    _pos += size;
    return size;
}

int64_t MemBuf::write(const char *buffer, int64_t size, int64_t offset) {
    if (_rw_strategy == ERwStrategy::S_RANDOM_RW) {
        _pos = offset;
        size = remain_size(size);
        if (_pos + size > _file_size) {
            _file_size = _pos + size;
        }
    } else if (_rw_strategy == ERwStrategy::S_MEM_SEQUENCE_RW) {
        if (!check_sequence(EMemCacheOp::OP_WRITE ,offset)) {
             BOSFS_ERR("write is not sequential. last op:%ld offset:%ld, size:%ld, cur offset:%ld",
             _last_op, _last_op_pos, _last_op_size, offset);
             return -1;
        }
        size = remain_size(size);
        _file_size += size;
        update_last_op(EMemCacheOp::OP_WRITE, offset, size);
    }
    memcpy((char *)_buffer + _pos, buffer, (size_t)size);
    _pos += size;
    return size;
}


int MemBuf::adjust_for_load(uint64_t need_load_size, int64_t offset) {
    BOSFS_DEBUG("before load adjust: _size:%ld, _pos:%ld, _data_pos:%ld, _pre_load_pos:%ld, offset:%ld",
            _size, _pos, _data_pos, _pre_load_pos, offset);

    if (_rw_strategy == ERwStrategy::S_RANDOM_RW) {
        // just do all data mapping
        int64_t cur_remain_load_size = remain_load_size();
        if (need_load_size > cur_remain_load_size) {
            BOSFS_ERR("need_load_size > remain_load_size in random mode. buffer_size:%ld, remain_load_size:%ld, need_load_size:%ld",
                       _size, cur_remain_load_size, need_load_size);
            return -1;
        }
        BOSFS_DEBUG("random model load data, buffer_size:%ld, load_offset:%ld, need_load_size:%ld ",
                     _size, offset, need_load_size);

    } else if (_rw_strategy == ERwStrategy::S_MEM_SEQUENCE_RW) {
        // need move unread data to front
        if (_pos + need_load_size > _size) {
            uint64_t need_move_size = _pre_load_pos - _pos;
            if (need_move_size >= 0) {
                memcpy(_buffer, (const char *)_buffer + _pos, need_move_size);
                _pos = 0;
                _data_pos = need_move_size;
                _pre_load_pos = _data_pos + need_load_size;
                if (_pre_load_pos > _size) {
                    BOSFS_ERR("_pre_load_pos is over buffer _size:%ld, _pos:%ld, _pre_load_pos:%ld", _size, _pos, _pre_load_pos);
                    return -1;
                }
                BOSFS_DEBUG("memove size:%ld, buffer_size:%ld, after _pos:%ld, after _data_pos:%ld, after _pre_load_pos:%ld",
                             need_move_size, _size, _pos, _data_pos, _pre_load_pos);
            } else {
                BOSFS_ERR("_pre_load_pos <  _pos buffer_size:%ld, _pos:%ld, _pre_load_pos:%ld", _size, _pos, _pre_load_pos);
                return -1;
            }
        } else {
            _data_pos = _pre_load_pos;
            _pre_load_pos = _data_pos + need_load_size;
            BOSFS_DEBUG("do not memove, buffer_size:%ld, after _pos:%ld, after _data_pos:%ld, after _pre_load_pos:%ld",
                        _size, _pos, _data_pos, _pre_load_pos);
        }
    }
    return 0;
}

END_FS_NAMESPACE
