/***************************************************************************
 *
 * Copyright (c) 2017 Baidu.com, Inc. All Rights Reserved
 *
 **************************************************************************/
/**
 * @file src/file_manager.cpp
 * @author lili19(com@baidu.com)
 * @date 2017/09/12 09:35:27
 * @brief
 *
 **/

#include <string.h>
#include "object_page.h"
#include "data_cache.h"
#include "file_manager.h"

BEGIN_FS_NAMESPACE

// Definition of class ObjectPageList
void ObjectPageList::free_list(self_type &lst) {
    for (self_type::iterator iter = lst.begin(); iter != lst.end(); iter = lst.erase(iter)) {
        delete *iter;
    }
    lst.clear();
}

ObjectPageList::ObjectPageList(size_t size, bool loaded) {
    init(size, loaded);
}

ObjectPageList::~ObjectPageList() {
    clear();
}

bool ObjectPageList::init(size_t size, bool loaded) {
    clear();
    ObjectPage *page = new ObjectPage(0, size, loaded);
    _pages.push_back(page);
    return true;
}

size_t ObjectPageList::get_size() const {
    if (_pages.empty()) {
        return 0;
    }
    self_type::const_reverse_iterator rite = _pages.rbegin();
    return static_cast<size_t>((*rite)->next());
}

bool ObjectPageList::resize(size_t size, bool loaded) {
    size_t total = get_size();

    if (0 == total) {
        init(size, loaded);
    } else if (total < size) {
        ObjectPage *page = new ObjectPage(static_cast<off_t>(total), (size - total), loaded);
        _pages.push_back(page);
    } else if (size < total) {
        for (self_type::iterator iter = _pages.begin(); iter != _pages.end();) {
            if (static_cast<size_t>((*iter)->next()) <= size) {
                ++iter;
            } else {
                if (size <= static_cast<size_t>((*iter)->get_offset())) {
                    delete *iter;
                    iter = _pages.erase(iter);
                } else {
                    (*iter)->set_bytes(size - static_cast<size_t>((*iter)->get_offset()));
                }
            }
        }
    } else {
        return false;
    }
    return compress();
}

bool ObjectPageList::is_page_loaded(off_t start, size_t size) const {
    for (self_type::const_iterator ite = _pages.begin(); ite != _pages.end(); ++ite) {
        if ((*ite)->end() < start) {
            continue;
        }
        if (!(*ite)->get_loaded()) {
            return false;
        }
        if (0 != size && static_cast<size_t>(start + size) <= static_cast<size_t>((*ite)->next())) {
            break;
        }
    }
    return true;
}

bool ObjectPageList::set_page_loaded_status(off_t start, size_t size, bool loaded, bool need_cmp) {
    size_t now_size = get_size();

    if (now_size <= static_cast<size_t>(start)) {
        if (now_size < static_cast<size_t>(start)) {
            resize(static_cast<size_t>(start), false);
        }
        resize(static_cast<size_t>(start + size), loaded);
    } else if (now_size <= static_cast<size_t>(start + size)) {
        resize(static_cast<size_t>(start), false);
        resize(static_cast<size_t>(start + size), loaded);
    } else {
        parse(start);
        parse(start + size);

        for (self_type::iterator iter = _pages.begin(); iter != _pages.end(); ++iter) {
            if ((*iter)->end() < start) {
                continue;
            } else if (static_cast<off_t>(start + size) <= (*iter)->get_offset()) {
                break;
            } else {
                (*iter)->set_loaded(loaded);
            }
        }
    }
    return need_cmp ? compress() : true;
}

bool ObjectPageList::find_unloaded_pate(off_t start, off_t &ret_start, size_t &ret_size) const {
    for (self_type::const_iterator iter = _pages.begin(); iter != _pages.end(); ++iter) {
        if (start < (*iter)->end()) {
            if (!(*iter)->get_loaded()) {
                ret_start = (*iter)->get_offset();
                ret_size = (*iter)->get_bytes();
                return true;
            }
        }
    }
    return false;
}

size_t ObjectPageList::get_total_unloaded_page_size(off_t start, size_t size) const {
    size_t ret_size = 0;
    off_t next = static_cast<off_t>(start + size);
    for (self_type::const_iterator iter = _pages.begin(); iter != _pages.end(); ++iter) {
        if ((*iter)->next() <= start) {
            continue;
        }
        if (next <= (*iter)->get_offset()) {
            break;
        }
        if ((*iter)->get_loaded()) {
            continue;
        }
        size_t tmp_size = 0;
        if ((*iter)->get_offset() <= start) {
            if ((*iter)->next() <= next) {
                tmp_size = static_cast<size_t>((*iter)->next() - start);
            } else {
                tmp_size = static_cast<size_t>(next - start);
            }
        } else {
            if ((*iter)->next() <= next) {
                tmp_size = static_cast<size_t>((*iter)->next() - (*iter)->get_offset());
            } else {
                tmp_size = static_cast<size_t>(next - (*iter)->get_offset());
            }
        }
        ret_size += tmp_size;
    }
    return ret_size;
}

int ObjectPageList::get_unloaded_pages(self_type &unloaded_list, off_t start, size_t size) const {
    if (0 == size) {
        size_t real_size = get_size();
        if (static_cast<size_t>(start) < real_size) {
            size = real_size - start;
        }
    }
    off_t next = static_cast<off_t>(start + size);
    for (self_type::const_iterator iter = _pages.begin(); iter != _pages.end(); ++iter) {
        if ((*iter)->next() <= start) {
            continue;
        }
        if (next <= (*iter)->get_offset()) {
            break;
        }
        if ((*iter)->get_loaded()) {
            continue;
        }

        off_t page_start = std::max((*iter)->get_offset(), start);
        off_t page_next = std::min((*iter)->next(), next);
        size_t page_size = static_cast<size_t>(page_next - page_start);

        self_type::reverse_iterator rite = unloaded_list.rbegin();
        if (rite != unloaded_list.rend() && (*rite)->next() == page_start) {
            (*rite)->set_bytes((*rite)->get_bytes() + page_size);
        } else {
            ObjectPage *page = new ObjectPage(page_start, page_size, false);
            unloaded_list.push_back(page);
        }
    }

    return unloaded_list.size();
}

bool ObjectPageList::serialize(StatCacheFile &file, bool is_output) {   
    if (!file.open_file()) {
        return false;
    }
    if (is_output) { 
        std::stringstream ssall;
        ssall << get_size();

        FilePtr filePtr;
        if (FileManager::instance().get(file.get_path(), &filePtr)) {
            BOSFS_ERR("FileManager failed to get file, %s", file.get_path().c_str());
            return false;
        }
        ssall << "\n";
        ssall << filePtr->last_modified_time();

        for (self_type::iterator iter = _pages.begin(); iter != _pages.end(); ++iter) {
            ssall << "\n"
                  << (*iter)->get_offset() << ":" << (*iter)->get_bytes() << ":"
                  << ((*iter)->get_loaded() ? "1" : "0");
        }

        std::string strall = ssall.str();
        // ftruncate before pwrite
        if (ftruncate(file.get_fd(), 0)) {
            BOSFS_ERR("failed to ftruncate. stats(%d)", errno);
            return false; 
        }
        if (0 >= pwrite(file.get_fd(), strall.c_str(), strall.length(), 0)) {
            BOSFS_ERR("failed to write. stats(%d)", errno);
            return false;
        }
        BOSFS_DEBUG("finish serialize object page info to .status file.");
        dump();

    } else { 
        BOSFS_DEBUG("ready serialize .status file to object page info.");
        struct stat st;
        memset(&st, 0, sizeof(struct stat));
        if (-1 == fstat(file.get_fd(), &st)) {
            BOSFS_ERR("fstat is failed. errno(%d)", errno);
            return false;
        }
        if (0 >= st.st_size) {
            init(0, false);
            return true;
        }
        SmartArray<char> tmp(st.st_size + 1);
        char *ptmp = tmp.get();

        if (0 >= pread(file.get_fd(), ptmp, st.st_size, 0)) {
            BOSFS_ERR("failed to read stats(%d)", errno);
            return false;
        }
        std::string oneline;
        std::stringstream ssall(ptmp);

        clear();

        //get total size
        if (!std::getline(ssall, oneline, '\n')) {
            BOSFS_ERR("failed to parse stats.");
            return false;
        }
        size_t total = strtoll(oneline.c_str(), NULL, 10);

        // get && check old modified time
        if (!std::getline(ssall, oneline, '\n')) {
            BOSFS_ERR("failed to parse stats.");
            return false;
        }
        size_t old_modified_time = strtoll(oneline.c_str(), NULL, 10);
        FilePtr filePtr;
        if (FileManager::instance().get(file.get_path(), &filePtr)) {
            BOSFS_ERR("FileManager failed to get file, %s", file.get_path().c_str());
            return false;
        }
        int64_t new_modified_time = filePtr->last_modified_time();
        if (new_modified_time > old_modified_time) {
            BOSFS_WARN("stats meta is expired. now reload. old:%ld, new:%ld", old_modified_time, new_modified_time);
            clear();
            return false;
        }
        BOSFS_DEBUG("stats file meta. old:%ld, new:%ld", old_modified_time, new_modified_time);

        bool is_err = false;
        while (std::getline(ssall, oneline, '\n')) {
            std::string part;
            std::stringstream ssparts(oneline);

            if (!std::getline(ssparts, part, ':')) {
                is_err = true;
                break;
            }
            off_t offset = strtoll(part.c_str(), NULL, 10);

            if (!std::getline(ssparts, part, ':')) {
                is_err = true;
                break;
            }
            off_t size = strtoll(part.c_str(), NULL, 10);
            if (!std::getline(ssparts, part, ':')) {
                is_err = true;
                break;
            }
            bool is_loaded = strtoll(part.c_str(), NULL, 10) != 0;
            set_page_loaded_status(offset, size, is_loaded);
        }
        if (is_err) {
            BOSFS_ERR("failed to parse stats.");
            clear();
            return false;
        }

        if (total != get_size()) {
            BOSFS_ERR("different size(%jd - %jd).", (intmax_t)total, (intmax_t)get_size());
            clear();
            return false;
        }
        dump();
    } 
    return true;
}

void ObjectPageList::dump() {
    int cnt = 0;
    std::ostringstream oss;
    oss << "ObjectPageList pages = [";
    for (self_type::iterator it = _pages.begin(); it != _pages.end(); ++it) {
        ObjectPage *page = *it;
        if (cnt > 0) {
            oss << "->";
        }
        oss << "(off=" << page->get_offset() << ",size=" << page->get_bytes()
            << ",load=" << page->get_loaded() << ")";
        ++cnt;
    }
    oss << "]";
    BOSFS_DEBUG("%s", oss.str().c_str());
}

void ObjectPageList::clear() {
    ObjectPageList::free_list(_pages);
}

//合并page
bool ObjectPageList::compress() {
    bool is_first = true;
    bool is_last_loaded = false;
    for (self_type::iterator iter = _pages.begin(); iter != _pages.end();) {
        if (is_first) {
            is_first = false;
            is_last_loaded = (*iter)->get_loaded();
            ++iter;
        } else {
            if (is_last_loaded == (*iter)->get_loaded()) {
                self_type::iterator bite = iter;
                --bite;
                (*bite)->set_bytes((*bite)->get_bytes() + (*iter)->get_bytes());
                delete *iter;
                iter = _pages.erase(iter);
            } else {
                is_last_loaded = (*iter)->get_loaded();
                ++iter;
            }
        }
    }

    return true;
}

bool ObjectPageList::parse(off_t new_pos) {
    for (self_type::iterator iter = _pages.begin(); iter != _pages.end(); ++iter) {
        if (new_pos == (*iter)->get_offset()) {
            return true;
        } else if ((*iter)->get_offset() < new_pos && new_pos < (*iter)->next()) {
            ObjectPage *page = new ObjectPage(
                    (*iter)->get_offset(),
                    static_cast<size_t>(new_pos - (*iter)->get_offset()),
                    (*iter)->get_loaded());
            (*iter)->set_bytes((*iter)->get_bytes() - new_pos + (*iter)->get_offset());
            (*iter)->set_offset(new_pos);
            _pages.insert(iter, page);
            return true;
        }
    }
    return false;
}

END_FS_NAMESPACE
